"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const constants_1 = require("./constants");
function format(options) {
    class Formatter {
        constructor(object) {
            this.keyword = 'type ';
            this.separator = ' ';
            this.body = '= any';
            this.description = '';
            this.deprecated = '';
            this.defaultValue = '';
            this.object = object;
            if (object.description)
                this.description = object.description;
            if (object.isDeprecated) {
                this.deprecated = '@deprecated';
                if (object.deprecationReason)
                    this.deprecated += ` ${object.deprecationReason}`;
            }
        }
        get Scalar() {
            const { name } = this.object;
            if (name in constants_1.default.scalars)
                this.body = `= ${constants_1.default.scalars[name]}`;
            return this.result;
        }
        get String() {
            this.separator = '';
            this.keyword = '\'';
            this.body = '\'';
            return this.result;
        }
        get Enum() {
            this.body = `=${this.object.enumValues.map(Formatter.String).join(' |')}`;
            return this.result;
        }
        get Union() {
            this.body = `= ${this.object.possibleTypes.map(object => object.name).join(' | ')}`;
            return this.result;
        }
        get Interface() {
            const { fields } = this.object;
            if (fields) {
                const extension = [];
                if (options.includeArguments)
                    extension.push({
                        name: '__arguments',
                        fields: fields
                            .filter(field => field.args.length)
                            .map(field => ({
                            name: field.name,
                            fields: field.args.map(object => (Object.assign({}, object, { _optional: true })))
                        }))
                    });
                if (options.includeScalars || options.includeObjects) {
                    const [scalars, objects] = partitionByKind(fields, constants_1.default.objectKinds);
                    if (options.includeScalars)
                        extension.push({
                            name: '__scalars',
                            fields: scalars.map(({ name, type, ofType }) => ({ name, type, ofType, _onlyPureType: true }))
                        });
                    if (options.includeObjects)
                        extension.push({
                            name: '__objects',
                            fields: objects.map(({ name, type, ofType }) => ({ name, type, ofType, _onlyPureType: true }))
                        });
                }
                fields.unshift(...extension);
            }
            this.keyword = 'interface ';
            return this.describe.result;
        }
        get Object() {
            const { interfaces } = this.object;
            if (interfaces && interfaces.length)
                this.separator = ` extends ${interfaces.map(object => object.name).join(', ')} `;
            return this.Interface;
        }
        get Property() {
            this.describe.keyword = '';
            this.separator = this.body.endsWith(' | null') && this.object._optional ? '?: ' : ': ';
            return this.result;
        }
        get describe() {
            const { object } = this;
            const fields = object.inputFields || object.fields;
            this.body = fields
                ? fields.length
                    ? `{${fields.map(Formatter.Property).join('')}\n}`
                    : '{}'
                : formatType(object.type, object._onlyPureType);
            return this.default;
        }
        get default() {
            const { defaultValue } = this.object;
            if (defaultValue !== null && defaultValue !== undefined)
                this.defaultValue = `@default ${defaultValue}`;
            return this;
        }
        get result() {
            let { name } = this.object;
            let comment = [this.description, this.deprecated, this.defaultValue]
                .filter(x => !!x)
                .map(string => string.replace(/\s*\n\s*/g, '\n * '))
                .join('\n * ');
            if (name.startsWith('__'))
                name = `$${name}`;
            if (comment)
                comment = `/**\n * ${comment}\n */\n`;
            return `\n${comment}${this.keyword}${name}${this.separator}${this.body}`.replace(/\n/g, '\n\t');
        }
    }
    Formatter.Scalar = object => new Formatter(object).Scalar;
    Formatter.String = object => new Formatter(object).String;
    Formatter.Enum = object => new Formatter(object).Enum;
    Formatter.Union = object => new Formatter(object).Union;
    Formatter.Interface = object => new Formatter(object).Interface;
    Formatter.Object = object => new Formatter(object).Object;
    Formatter.Property = object => new Formatter(object).Property;
    const typeMapping = {
        SCALAR: Formatter.Scalar,
        ENUM: Formatter.Enum,
        UNION: Formatter.Union,
        INPUT_OBJECT: Formatter.Object,
        INTERFACE: Formatter.Interface,
        OBJECT: Formatter.Object
    };
    return json => {
        const schema = json.data.__schema;
        const { types } = schema;
        const dictionary = types
            .map(({ name }) => Formatter.Property({ name, _onlyPureType: true, type: { name } }))
            .join('');
        const dictionaries = types.reduce((dictionaries, object) => {
            const { kind } = object;
            const value = typeMapping[kind](object);
            if (kind in dictionaries)
                dictionaries[kind].push(value);
            else
                dictionaries[kind] = [value];
            return dictionaries;
        }, {});
        const typeDefinitions = Object
            .keys(dictionaries)
            .map(key => dictionaries[key].join('\n'))
            .join('\n\n');
        const primitives = Object
            .keys(constants_1.default.primitives)
            .map(primitive => {
            const key = constants_1.default.primitives[primitive];
            return `export type ${primitive} = ${schema[key]
                ? `${constants_1.default.namespace}.${schema[key].name}`
                : 'null'}`;
        })
            .join('\n');
        return `declare namespace ${constants_1.default.namespace} {${typeDefinitions}\n}\n\n${primitives}\n\nexport interface Types {\n\t$__queryType: Query\n\t$__mutationType: Mutation\n\t$__subscriptionType: Subscription${dictionary}\n}\n\nexport default Types\n`;
    };
}
exports.format = format;
function ofKind(type, kinds) {
    switch (type.kind) {
        case 'NON_NULL':
        case 'LIST': return ofKind(type.ofType, kinds);
        default: return kinds.includes(type.kind);
    }
}
function partitionByKind(items, kinds) {
    return items.reduce((items, item) => {
        items[+ofKind(item.type, kinds)].push(item);
        return items;
    }, [[], []]);
}
function formatType(type, onlyPureType = false) {
    switch (type.kind) {
        case 'LIST': {
            const result = formatType(type.ofType, onlyPureType);
            return onlyPureType ? result : `Array<${result}>`;
        }
        case 'NON_NULL': {
            const result = formatType(type.ofType, onlyPureType);
            return onlyPureType ? result : result.replace(' | null', '');
        }
        case 'OBJECT':
        case 'INTERFACE':
        case 'SCALAR':
        case 'ENUM':
        case 'UNION':
        case 'INPUT_OBJECT':
        default: {
            const prefix = type.name.startsWith('__') ? '$' : '';
            const suffix = onlyPureType ? '' : ' | null';
            return `${constants_1.default.namespace}.${prefix}${type.name}${suffix}`;
        }
    }
}
exports.default = format;
